import functools
import os
import sys
import unittest
import examples

PATH = os.path.abspath(os.path.dirname(__file__))
sys.path.append(os.path.join(PATH, os.path.pardir, "scangen"))

from rnd import ExprSymbols, convert, Dfa, DfaSymbols
from rnd.internals import crnd

TEST_DATA = {}

def sym(a=None, b=None):
    if a is None:
        return ExprSymbols()
    return ExprSymbols(ord(a), None if b is None else ord(b))

def closed_range(a, b=None):
    a = ord(a)
    if b:
        b = ord(b)+1
    else:
        b = min(0xffffffff, a+1)
    return ExprSymbols(a, b)

def optional(expr):
    return expr.union(sym())

def to_dfa(expr):
    """Converts expr to dfa and destroys expr."""
    dfa = convert(expr)
    expr.destroy()
    return dfa

def empty_expr():
    return sym()

def identifier_expr():
    letters = closed_range('_').union(closed_range('a', 'z')).union(closed_range('A', 'Z'))
    return letters.concatenation(letters.union(closed_range('0', '9')).closure())

def integer_expr():
    return closed_range('0').union(closed_range('1', '9').concatenation(closed_range('0', '9').closure()))

def whitespace_expr():
    return closed_range(' ').union(closed_range('\t')).union(closed_range('\n'))

def number_expr():
    digit = closed_range('0', '9')
    decimal = closed_range('.').concatenation(digit).concatenation(digit.closure())
    sign = optional(closed_range('-').union(closed_range('+')))
    exponent = (closed_range('e').union(closed_range('E'))).concatenation(sign).concatenation(integer_expr())
    return integer_expr().concatenation(optional(decimal)).concatenation(optional(exponent))

def string_expr():
    char = functools.reduce(lambda a, b: a.union(b),
            [ ExprSymbols(32, 33+1), ExprSymbols(35, 91+1), ExprSymbols(93, 126+1) ])
    char = char.union(closed_range("\\").concatenation(ExprSymbols(32, 126+1)))
    string = char.closure()
    return closed_range('"').concatenation(string).concatenation(closed_range('"'))

def character_expr():
    escape = closed_range("\\").concatenation(functools.reduce(lambda a, b: a.union(b),
        [closed_range("'"), closed_range("\\"), closed_range("t"), closed_range("n")]))
    middle = functools.reduce(lambda a, b: a.union(b),
            [ ExprSymbols(32, 38+1), ExprSymbols(40, 91+1), ExprSymbols(93, 126+1),
                escape ])
    return closed_range("'").concatenation(middle).concatenation(closed_range("'"))

EXPR_FN = {
    "empty": empty_expr,
    "number": number_expr,
    "identifier": identifier_expr,
    "string": string_expr,
    "character": character_expr,
    "whitespace": whitespace_expr,
}

def parametrize_expr_test(test_case, expr_type):
    test_case.dfa = to_dfa(EXPR_FN[expr_type]())
    test_case.assertTrue(TEST_DATA[expr_type])
    for word, label in TEST_DATA[expr_type]:
        word = word.encode().decode()
        accepted = test_case.dfa.compute(map(ord, word))
        label = bool(label)
        test_case.assertEqual(accepted, label)

class RndConversionTest(unittest.TestCase):
    def setUp(self):
        self.dfa = Dfa()

    def tearDown(self):
        self.dfa = Dfa()

    def test_leaks(self):
        for expr_fn in EXPR_FN.values():
            to_dfa(expr_fn())
        self.assertEqual(0, crnd.rnd_get_expr_counter())

    def test_random_data(self):
        for expr_type in EXPR_FN:
            with self.subTest(expr_type=expr_type):
                parametrize_expr_test(self, expr_type)

if __name__ == "__main__":
    for expr_type in EXPR_FN:
        with open(f"{PATH}/data/{expr_type}.csv", "r") as file:
            TEST_DATA[expr_type] = examples.read(file)
    unittest.main()



"""Direct wrapper for libcrnd."""

import ctypes

crnd = ctypes.CDLL("librndpoc.so")

class CRange(ctypes.Structure):
    """Wrapper for struct rnd_range."""

    _fields_ = [
        ("start", ctypes.c_uint32),
        ("end", ctypes.c_uint32)
    ]

class CExpr(ctypes.Structure):
    """Wrapper for struct rnd_expr."""

CExpr._fields_ = [
    ("type", ctypes.c_char),
    ("value", CRange),
    ("left", ctypes.POINTER(CExpr)),
    ("right", ctypes.POINTER(CExpr))
]

class CTransition(ctypes.Structure):
    """Wrapper for struct rnd_transition."""

    _fields_ = [
        ("label", CRange),
        ("to", ctypes.c_int)
    ]

class CState(ctypes.Structure):
    """Wrapper for struct rnd_state."""

    _fields_ = [
        ("accept", ctypes.c_bool),
        ("outdegree", ctypes.c_int),
        ("transitions", ctypes.POINTER(CTransition))
    ]

class CDfa(ctypes.Structure):
    """Wrapper for struct rnd_dfa."""

    _fields_ = [
        ("order", ctypes.c_int),
        ("states", ctypes.POINTER(CState)),
        ("error", ctypes.c_int)
    ]

crnd.rnd_convert.argtypes = [ctypes.POINTER(CExpr)]
crnd.rnd_convert.restype = CDfa

crnd.rnd_dfa_destroy.argtypes = [ctypes.POINTER(CDfa)]
crnd.rnd_dfa_destroy.restype = None

crnd.rnd_symbol.argtypes = [ctypes.c_uint32, ctypes.c_uint32]
crnd.rnd_symbol.restype = ctypes.POINTER(CExpr)

crnd.rnd_union.argtypes = [ctypes.POINTER(CExpr), ctypes.POINTER(CExpr)]
crnd.rnd_union.restype = ctypes.POINTER(CExpr)

crnd.rnd_concatenation.argtypes = [ctypes.POINTER(CExpr), ctypes.POINTER(CExpr)]
crnd.rnd_concatenation.restype = ctypes.POINTER(CExpr)

crnd.rnd_closure.argtypes = [ctypes.POINTER(CExpr)]
crnd.rnd_closure.restype = ctypes.POINTER(CExpr)

crnd.rnd_expr_destroy.argtypes = [ctypes.POINTER(CExpr)]
crnd.rnd_expr_destroy.restype = None

crnd.rnd_expr_free.argtypes = [ctypes.POINTER(CExpr)]
crnd.rnd_expr_free.restype = None

crnd.rnd_get_expr_counter.restype = ctypes.c_int



"""Wrapper for libcrnd with additional features."""

import ctypes
import enum
import os
import sys
from redblack import containers
from .internals import crnd

class ExprType(enum.Enum):
    """Same as enum rnd_type { RND_SYMBOL, RND_UNION, ... }."""
    SYMBOL = 0
    UNION = 1
    CONCATENATION = 2
    CLOSURE = 3

class ExprSymbols:
    """Class that represents a closed range of symbols.

    Needs to be destroyed after use, either directly or indirectly.
    """

    def __init__(self, start=0xffffffff, end=None):
        if end is None:
            end = min(0xffffffff, start+1)
        if start > end:
            raise ValueError("start must be <= end")
        self.start = start
        self.end = end
        self.cpointer = crnd.rnd_symbol(start, end)

    def union(self, other):
        """Return a UNION Expr.

        other may be an ExprSymbols or an Expr.
        """
        return union(self, other)

    def concatenation(self, other):
        """Return a CONCATENATION Expr.

        other may be an ExprSymbols or an Expr.
        """
        return concatenation(self, other)

    def closure(self):
        """Return a CLOSURE Expr."""
        return closure(self)

    def destroy(self):
        """Free memory allocated during __init__."""
        if self.cpointer:
            crnd.rnd_expr_free(self.cpointer)
        self.cpointer = None

    def __repr__(self):
        if self.start >= self.end:
            return "nil"
        elif self.start+1 == self.end:
            return str(self.start)
        return f"[{self.start}, {self.end})"

def _get_expr_pointer(expr):
    return None if not expr else expr.cpointer

# doesn't raise exceptions to avoid memory leak
class Expr:
    """Owner class of CExpr with methods for creating new Expr.

    Needs to be destroyed after use, either directly or indirectly.
    """

    def __init__(self, type_, left=None, right=None):
        self.type_ = type_
        self.left = left
        self.right = right
        self.cpointer = None

        left = _get_expr_pointer(self.left)
        if type_ is ExprType.UNION:
            self.cpointer = crnd.rnd_union(
                left,
                _get_expr_pointer(self.right))
        elif type_ is ExprType.CONCATENATION:
            self.cpointer = crnd.rnd_concatenation(
                left,
                _get_expr_pointer(self.right))
        elif type_ is ExprType.CLOSURE:
            self.cpointer = crnd.rnd_closure(left)

    def __repr__(self):
        if self.type_ == ExprType.UNION:
            return f"union({self.left!r}, {self.right!r})"
        elif self.type_ == ExprType.CONCATENATION:
            return f"concatenation({self.left!r}, {self.right!r})"
        elif self.type_ == ExprType.CLOSURE:
            return f"closure({self.left!r})"
        assert False

    def destroy(self):
        """Free cpointer of this node and all descendant nodes.

        Does not need to be called if invoked from another Expr.
        """
        if self.cpointer:
            crnd.rnd_expr_free(self.cpointer)
        self.cpointer = None
        if self.left:
            self.left.destroy()
        if self.right:
            self.right.destroy()
        self.left = None
        self.right = None

    def union(self, other):
        """Return a UNION Expr."""
        return union(self, other)

    def concatenation(self, other):
        """Return a CONCATENATION Expr."""
        return concatenation(self, other)

    def closure(self):
        """Return a CLOSURE Expr."""
        return closure(self)

def union(a, b) -> Expr:
    return Expr(ExprType.UNION, a, b)

def concatenation(a, b) -> Expr:
    return Expr(ExprType.CONCATENATION, a, b)

def closure(expr: Expr or ExprSymbols) -> Expr:
    return Expr(ExprType.CLOSURE, expr)

class DfaSymbols:
    """Similar to CRange, but with built in comparator."""

    def __init__(self, start=0xffffffff, end=None):
        if end is None:
            end = min(0xffffffff, start+1)
        if start > end:
            raise ValueError("start must be <= end")
        self.start = start
        self.end = end

    def __repr__(self):
        if self.start >= self.end:
            return "nil"
        elif self.start+1 == self.end:
            return str(self.start)
        return f"[{self.start}, {self.end})"

    def __lt__(self, other):
        """Check if self is on the left of other."""
        assert self.start <= self.end
        assert other.start <= other.end
        return self.end <= other.start

    def __gt__(self, other):
        """Check if self is on the right of other."""
        return other < self

    def __eq__(self, other):
        """Check if self and other overlap."""
        return not (self < other) and not (self > other)

class Dfa:
    """Higher-level wrapper for rnd_dfa with computational capabilities."""

    def __init__(self):
        self.start = 0
        self.accepts = set()
        self.transitions = {}
        self.error = -1

    def __repr__(self):
        return f"<rnd.Dfa start={self.start!r}, accepts={self.accepts!r}, "\
                f"error={self.error}, "\
                f"transitions={self.transitions!r}>"

    def compute(self, inputs):
        """Compute if Dfa accepts the sequence of ints.

        Assume that -1 is an error state with no outbound transitions.
        """
        state = self.start
        for a in inputs:
            if state not in self.transitions or state == self.error:
                return False
            state = self.transitions[state].get(DfaSymbols(a), -1)
        return state in self.accepts

def _cdfa_to_pydfa(_dfa: internals.CDfa) -> Dfa:
    dfa = Dfa()
    dfa.transitions[dfa.start] = containers.Map()
    dfa.error = _dfa.error

    for q in range(_dfa.order):
        state = _dfa.states[q]
        if q not in dfa.transitions:
            dfa.transitions[q] = containers.Map()
        if state.accept:
            dfa.accepts.add(q)
        for j in range(state.outdegree):
            trans = state.transitions[j]
            a = DfaSymbols(int(trans.label.start), int(trans.label.end))
            r = int(trans.to)
            if r not in dfa.transitions:
                dfa.transitions[r] = containers.Map()
            # assumes a doesn't overlap with any other DfaSymbols
            dfa.transitions[q][a] = r
    return dfa

def convert(expr: Expr or ExprSymbols) -> Dfa:
    """Convert expr into dfa, without destroying expr."""
    _dfa = crnd.rnd_convert(expr.cpointer)
    dfa = _cdfa_to_pydfa(_dfa)
    crnd.rnd_dfa_destroy(ctypes.byref(_dfa))
    return dfa

name = "rnd"



"""A scanner generator that uses Jinja2 templates."""
import argparse
import sys
import jinja2
from . import rnd

scanners = []

def symbols(left: str = None, right: str = None) -> rnd.ExprSymbols:
    """Create range of symbols from characters."""
    left = ord(left) if left is not None else 0xffffffff
    if right:
        right = ord(right)+1
    else:
        right = min(0xffffffff, left+1)
    return rnd.ExprSymbols(left, right)

def isymbols(left: int = None, right: int = None) -> rnd.ExprSymbols:
    """Create range of symbols from integers."""
    if left is None:
        left = 0xffffffff
    right = right+1 if right is not None else min(0xffffffff, left+1)
    return rnd.ExprSymbols(left, right)

def optional(expr: rnd.ExprSymbols or rnd.Expr):
    """Return union of expr with empty symbol."""
    return expr.union(symbols())

def token(name):
    """Register scanner."""
    def decorator(func):
        def wrapper():
            expr = func()
            dfa = rnd.convert(expr)
            expr.destroy()
            dfa.token = name
            return dfa
        scanners.append(wrapper())
        return wrapper
    return decorator

def generate(entrypoint="", directory=""):
    """Generate code from templates."""
    def get_args(args):
        description = "Generate scanner using jinja2 templates."
        parser = argparse.ArgumentParser(description=description)
        parser.add_argument("entrypoint", help="filename of template entrypoint")
        parser.add_argument("-d", default="", dest="directory",
                            help="path to templates directory")
        return parser.parse_args(args)

    def generate_code():
        loader = jinja2.FileSystemLoader(directory)
        env = jinja2.Environment(loader=loader)
        template = env.get_template(entrypoint)
        return template.render(scanners=scanners)

    args = sys.argv[1:]
    if entrypoint:
        args.append(entrypoint)
    if directory:
        args.extend(["-d", directory])
    args = get_args(args)
    if args.entrypoint:
        entrypoint = args.entrypoint
    if args.directory:
        directory = args.directory

    try:
        output = generate_code()
        print(output)
    except jinja2.exceptions.TemplateNotFound:
        print("scangen: Template not found:", entrypoint)

name = "scangen"



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="scangen",
    version="0.2a3",
    author="Levi Gruspe",
    author_email="mail.levig@gmail.com",
    description="Template-based scanner generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lggruspe/scangen",
    packages=["scangen", "scangen.rnd"],
    install_requires=["jinja2", "redblack"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux"
    ])
